@@ -1,10 +1,13 @@
+import deepdish as dd
+import h5py
 import os
 import xmltodict
 import yaml
 
 from typing import Dict, Any, List, Optional
-from pydantic import BaseModel, validate_arguments
+from pydantic import BaseModel, validate_arguments, Field
 from json import dumps
+from easyDataverse.core.file import File
 
 from easyDataverse.core.uploader import upload_to_dataverse, update_dataset
 from easyDataverse.core.dataverseBase import DataverseBase
@@ -13,19 +16,16 @@ from easyDataverse.tools.utils import YAMLDumper, get_class
 
 
 class Dataset(BaseModel):
-
     class Config:
         extra = "allow"
 
     metadatablocks: Dict[str, Any] = dict()
     p_id: Optional[str] = None
+    files: List[File] = Field(default_factory=list)
 
     # ! Adders
 
-    def add_metadatablock(
-        self,
-        metadatablock: DataverseBase
-    ) -> None:
+    def add_metadatablock(self, metadatablock: DataverseBase) -> None:
         """Adds a metadatablock object to the dataset if it is of 'DataverseBase' type and has a metadatablock name"""
 
         # Check if the metadatablock is of 'DataverseBase' type
@@ -34,34 +34,88 @@ class Dataset(BaseModel):
                 f"Expected class of type 'DataverseBase', got '{metadatablock.__class__.__name__}'"
             )
 
-        if hasattr(metadatablock, '_metadatablock_name') is False:
+        if hasattr(metadatablock, "_metadatablock_name") is False:
             raise TypeError(
                 f"The provided class {metadatablock.__class__.__name__} has no metadatablock name and is thus not compatible with this function."
             )
 
         # Add the metadatablock to the dataset as a dict
         block_name = getattr(metadatablock, "_metadatablock_name")
-        self.metadatablocks.update(
-            {block_name: metadatablock}
-        )
+        self.metadatablocks.update({block_name: metadatablock})
 
         # ... and to the __dict__
         setattr(self, block_name, metadatablock)
 
+    def add_file(self, dv_path: str, local_path: str, description: str = ""):
+        """Adds a file to the dataset based on the provided path.
+
+        Args:
+            filename (str): Path to the file to be added.
+            description (str, optional): Description of the file. Defaults to "".
+        """
+
+        # Create the file
+        file = File(filename=dv_path, local_path=local_path, description=description)
+
+        if file not in self.files:
+            self.files.append(file)
+        else:
+            raise Exception(f"File has already been added to the dataset")
+
+    def add_directory(self, dirpath: str) -> None:
+        """Adds a complete directory to the dataset.
+
+        Args:
+            dirpath (str): Path to the directory
+        """
+
+        dirpath = os.path.join(dirpath)
+
+        if not os.path.isdir(dirpath):
+            raise FileNotFoundError(
+                f"Directory at {dirpath} does not exist or is not a directory. Please provide a valid directory."
+            )
+
+        for path, _, files in os.walk(dirpath):
+
+            if "." in path:
+                continue
+
+            for file in files:
+                if file.startswith("."):
+                    continue
+
+                # Get all the metadata
+                filepath = os.path.join(path, file)
+
+                path_parts = [
+                    p
+                    for p in filepath.split(os.path.sep)
+                    if not p in dirpath.split(os.path.sep)
+                ]
+                filename = os.path.join(*path_parts)
+                data_file = File(filename=filename, local_path=filepath)
+
+                # Substitute new files with old files
+                found = False
+                for f in self.files:
+                    if f.filename == filename:
+                        f.local_path = data_file.local_path
+                        found = True
+                        break
+
+                if not found:
+                    self.files.append(data_file)
+
     # ! Exporters
 
     def xml(self) -> str:
         """Returns an XML representation of the dataverse object."""
 
         # Turn all keys to be camelcase
-        fields = self._keys_to_camel(
-            {"dataset_version": self.dict()}
-        )
+        fields = self._keys_to_camel({"dataset_version": self.dict()})
 
-        return xmltodict.unparse(
-            fields,
-            pretty=True, indent="    "
-        )
+        return xmltodict.unparse(fields, pretty=True, indent="    ")
 
     def dataverse_dict(self) -> dict:
         """Returns a dictionary representation of the dataverse dataset."""
@@ -71,11 +125,7 @@ class Dataset(BaseModel):
         for block in self.metadatablocks.values():
             blocks.update(block.dataverse_dict())
 
-        return {
-            "datasetVersion": {
-                "metadataBlocks": blocks
-            }
-        }
+        return {"datasetVersion": {"metadataBlocks": blocks}}
 
     def dataverse_json(self, indent: int = 2) -> str:
         """Returns a JSON representation of the dataverse dataset."""
@@ -103,25 +153,50 @@ class Dataset(BaseModel):
             data, Dumper=YAMLDumper, default_flow_style=False, sort_keys=False
         )
 
+    def hdf5(self, path: str) -> None:
+        """Exports the dataset to an HDF5 dataset that can also be read by the API_TOKEN
+
+        Args:
+            path (str): Path to the destination HDF5 files.
+        """
+
+        # Write metatdat to hdf5
+        dd.io.save(path, self.dict(exclude={"files"}))
+
+        # Add all files
+        with h5py.File(path, "r+") as f:
+
+            # Create Files Tag
+            files = f.create_group("Files")
+
+            for file in self.files:
+                if not file.local_path:
+                    # Skip if there is no actual data
+                    continue
+
+                # Setup metdata and content for hdf5 export
+                dset = files.create_group(file.filename.replace("/", "\\"))
+                dset.attrs["content"] = open(file.local_path, "rb").read()
+                if file.description:
+                    dset.attrs["description"] = file.description
+
     # ! Dataverse interfaces
 
     def upload(
         self,
         dataverse_name: str,
-        filenames: List[str] = None,
+        content_loc: Optional[str] = None,
         DATAVERSE_URL: Optional[str] = None,
-        API_TOKEN: Optional[str] = None
+        API_TOKEN: Optional[str] = None,
     ) -> str:
         """Uploads a given dataset to a Dataverse installation specified in the environment variable.
 
         Args:
             dataverse_name (str): Name of the target dataverse.
             filenames (List[str], optional): File or directory names which will be uploaded. Defaults to None.
+            content_loc (Optional[str], optional): If specified, the ZIP that is used to upload will be stored at the destination provided. Defaults to None.
             DATAVERSE_URL (Optional[str], optional): Alternatively provide base url as argument. Defaults to None.
             API_TOKEN (Optional[str], optional): Alternatively provide the api token as argument. Attention, do not use this for public scripts, otherwise it will expose your API Token. Defaults to None.
-            contact_name (Optional[str], optional): Provide a name that can be contacted for any requests regarding the dataset. Defaults to None.
-            contact_mail (Optional[str], optional): Provide an email that can be used to send requests regarding the dataset. Defaults to None.
-
         Returns:
             str: [description]
         """
@@ -129,10 +204,11 @@ class Dataset(BaseModel):
         self.p_id = upload_to_dataverse(
             json_data=self.dataverse_json(),
             dataverse_name=dataverse_name,
-            filenames=filenames,
+            files=self.files,
             p_id=self.p_id,
             DATAVERSE_URL=DATAVERSE_URL,
-            API_TOKEN=API_TOKEN
+            API_TOKEN=API_TOKEN,
+            content_loc=content_loc,
         )
 
         return self.p_id
@@ -141,9 +217,9 @@ class Dataset(BaseModel):
         self,
         contact_name: Optional[str] = None,
         contact_mail: Optional[str] = None,
-        filenames: Optional[list[str]] = None,
+        content_loc: Optional[str] = None,
         DATAVERSE_URL: Optional[str] = None,
-        API_TOKEN: Optional[str] = None
+        API_TOKEN: Optional[str] = None,
     ):
         """Updates a given dataset if a p_id has been given.
 
@@ -156,15 +232,14 @@ class Dataset(BaseModel):
         Args:
             contact_name (str, optional): Name of the contact. Defaults to None.
             contact_mail (str, optional): Mail of the contact. Defaults to None.
+            content_loc (Optional[str], optional): If specified, the ZIP that is used to upload will be stored at the destination provided. Defaults to None.
         """
 
         # Update contact
         if contact_name is None and contact_mail is None:
             # Check if there is already a contact defined
             contact_mails = [
-                contact.email
-                for contact in self.citation.contact
-                if contact.email
+                contact.email for contact in self.citation.contact if contact.email
             ]
 
             if len(contact_mails) == 0:
@@ -181,24 +256,22 @@ class Dataset(BaseModel):
 
         if has_mail == False:
             # Assign a completely new contact if the name is new
-            self.citation.add_contact(
-                name=contact_name,
-                email=contact_mail
-            )
+            self.citation.add_contact(name=contact_name, email=contact_mail)
 
         update_dataset(
             json_data=self.dataverse_dict()["datasetVersion"],
             p_id=self.p_id,
+            files=self.files,
             DATAVERSE_URL=DATAVERSE_URL,
             API_TOKEN=API_TOKEN,
-            filenames=filenames
+            content_loc=content_loc,
         )
 
     # ! Initializers
 
     @classmethod
     @validate_arguments
-    def from_dataverse_doi(cls, doi: str):
+    def from_dataverse_doi(cls, doi: str, filedir: str = "."):
         """Initializes a Dataset from a given Dataverse dataset.
 
         Args:
@@ -211,7 +284,7 @@ class Dataset(BaseModel):
 
         lib_name = os.environ["EASYDATAVERSE_LIB_NAME"]
 
-        return download_from_dataverse(cls(), doi, lib_name)
+        return download_from_dataverse(cls(), doi, lib_name, filedir)
 
     @classmethod
     def from_yaml(cls, path: str, use_id: bool = True):
@@ -235,11 +308,11 @@ class Dataset(BaseModel):
         Returns:
             Dataset: The resulting Dataset instance.
         """
-        
+
         # Load YAML file
-        with open(path, 'r') as file_handle:
+        with open(path, "r") as file_handle:
             data = yaml.safe_load(file_handle.read())
-        
+
         # Initialize blank dataset
         # and get lib_name for imports
         dataset = cls()
@@ -255,33 +328,29 @@ class Dataset(BaseModel):
 
             # Adapt module name to the namespace of generated code
             module_name = f".metadatablocks.{module_name}"
-            
-            # Retrieve class and initialize using the given 
+
+            # Retrieve class and initialize using the given
             # YAML data as a dicitonary
             cls = get_class(module_name, lib_name)[-1]
             instance = cls(**fields)
-            
+
             dataset.add_metadatablock(instance)
 
         return dataset
 
-        
-
     # ! Utilities
 
     @staticmethod
     def _snake_to_camel(word: str) -> str:
-        return ''.join(x.capitalize() or '_' for x in word.split('_'))
+        return "".join(x.capitalize() or "_" for x in word.split("_"))
 
     def _keys_to_camel(self, dictionary: dict):
         nu_dict = {}
         for key in dictionary.keys():
             if isinstance(dictionary[key], dict):
-                nu_dict[
-                    self._snake_to_camel(key)
-                ] = self._keys_to_camel(dictionary[key])
+                nu_dict[self._snake_to_camel(key)] = self._keys_to_camel(
+                    dictionary[key]
+                )
             else:
-                nu_dict[
-                    self._snake_to_camel(key)
-                ] = dictionary[key]
+                nu_dict[self._snake_to_camel(key)] = dictionary[key]
         return nu_dict
